/*
 * Copyright (c) 1990 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Lawrence Berkeley Laboratory,
 * Berkeley, CA.  The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
#ifndef lint
static char rcsid[] =
"@(#) $Header: aud.c,v 1.5 93/01/25 10:43:20 floyd Exp $ (LBL)";
#endif

#include <stdio.h>
#include <sys/types.h>
#include <math.h>

#include "config.h"
#include "real.h"
#include "packet.h"
#include "tcp.h"
#include "at.h"


struct audstate {
        int size;
        int interval;
        float ave_on;     /* average no. of intervals on */
        float ave_off;    /* average no. of intervals off */
        int started;
        int frame;	/* frame==1 for talk spurt */
        int count;      /* no. of pkts in talk spurt */
	int pkts;	/* number of packets so far */
	int maxpkts;            /* maximum number of packets */
	int max_burst;	/* max number of packets at peak rate */
	int bucket;	/* token bucket for enforcing the avgrate */
	int green;	/* =1 for generating green and red packets */
			/* red packets are excess packets */
	double last_token;	/* time of last token added to bucket */
	double avgrate;	/* avgrate as a fraction of the max rate */
};

/* exponential with mean f */
static inline int
rnd_exp(f)
        float f;
{
        return (int)(-log(1.0 - random() / 2147483647.0) * f);
}

/* geometric with mean f */
/* X = 1 + Floor(log U / log (1-1/u)), for U uniform [0,1], mean u. */
static inline int
rnd_geo(f)
        float f;
{
        return (1 + (int)(log(random() / 2147483647.0)/
                           log(1.0 - 1.0/(double)f ) ) );
}

static struct at_options options[] = {
        { "interval", AT_USEC, offsetof(struct audstate, interval) },
        { "size", AT_INT, offsetof(struct audstate, size) },
	{ "stop-at-packet", AT_INT, offsetof(struct audstate, maxpkts) },
	{ "max_burst", AT_INT, offsetof(struct audstate, max_burst) },
	{ "green", AT_INT, offsetof(struct audstate, green) },
	{ "avgrate", AT_DOUBLE, offsetof(struct audstate, avgrate) },
        { 0 },
        { 0 },
        { 0 },
        { 0 },
        { 0 },
};

void
aud_init(ap)
        register struct application *ap;
{
        register struct audstate *stp;

        stp = (struct audstate *)malloc(sizeof *stp);
        bzero((char *)stp, sizeof(*stp));
        atl_init(ap->ap_nb, ap->ap_attr, options, (char *)stp);
	ap->maxrate = (double) stp->size/(0.000001 * stp->interval);
	ap->avgrate = stp->avgrate * ap->maxrate;
	stp->pkts = 0;
        ap->ap_size = stp->size;
	ap->ap_type = PSTART;
        ap->ap_state = (void *)stp;
        stp->ave_on = (float) ( 352000.0 / ((float)stp->interval) );
        stp->ave_off = (float) (650000.0 / ((float) stp->interval) );
        stp->count = rnd_geo(stp->ave_on);
	stp->bucket = stp->max_burst;
	stp->last_token = 0.0;
}

int
aud_send(ap)
        register struct application *ap;
{
        struct audstate *stp = (struct audstate *)ap->ap_state;

        if (!stp->started) {
                stp->started = 1;
		stp->last_token = realtime();
                apptimer(ap->ap_proto->pr_cid, stp->interval);
        }
        return 0;
}

/* Aud_timeout sends at most "maxpkts" packets.  Before each packet is
 * sent, the token bucket is updated.  The max rate is enforced by the
 * timeout mechanism.
 * The number of new tokens is determined by the advertised average rate.
 * Packets that are "excess" are either dropped or marked "red" instead
 * of "green".
*/
void
aud_timeout(ap, id)
        struct application *ap;
        int id;
{
        struct audstate *stp = (struct audstate *)ap->ap_state;
	int new_tokens;
	new_tokens = (int) floor( (double) stp->avgrate * 
		1000000.0/stp->interval * (realtime() - stp->last_token) );
	if (new_tokens > 0) {
		stp->bucket += new_tokens;
		if (stp->bucket > stp->max_burst) stp->bucket = stp->max_burst;
		stp->last_token += (double) new_tokens /
			(stp->avgrate * 1000000.0/stp->interval);
	}
        if (stp->frame == 0) {
          stp->frame = 1;
          stp->count = rnd_geo(stp->ave_on);
        }
	if (stp->maxpkts==0 || stp->maxpkts > stp->pkts) {
	  if (stp->bucket > 0 || stp->green==1) {
	        --stp->count;
		++stp->pkts;
		if (stp->pkts == stp->maxpkts)
			ap->ap_type = PSTOP;
		else if (stp->pkts > 1) 
			ap->ap_type = AUDIO;
		if (stp->bucket > 0) {
			/* sent a green packet */
			--stp->bucket;
			ap->pri = 1;
	        	(*ap->ap_proto->pr_async)(ap->ap_proto, 1);
		}
		else {
			/* send a red packet */
			ap->pri = 2;
			(*ap->ap_proto->pr_async)(ap->ap_proto, 1);
		}
	        if (stp->count > 0)
	          apptimer(ap->ap_proto->pr_cid, stp->interval);
	        else {
	           stp->frame = 0;
	           apptimer(ap->ap_proto->pr_cid, 
			rnd_exp(stp->ave_off)*stp->interval );
	        }
	  }
	  else apptimer(ap->ap_proto->pr_cid, stp->interval);
	}
}
