/*
 * Copyright (c) 1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: object.cc,v 1.12 95/11/04 14:28:47 mccanne Exp $ (LBL)";
#endif

#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include "Tcl.h"
#include "object.h"
#include "trace.h"

class Var {
 public:
	Var(const char* name, int readonly);
	virtual ~Var();
	Var* next_;
	inline const char* name() const { return (name_); }
	virtual void set(const char*) = 0;
	virtual const char* get() = 0;
	inline int readonly() const { return (readonly_); }
 protected:
	char* name_;
	int readonly_;
	/*XXX make this static? */
	char wrk_[32];
};

class VarReal : public Var {
 public:
	VarReal(const char* name, int readonly, double* val);
	virtual void set(const char*);
	virtual const char* get();
 protected:
	double* val_;
};

class VarBandwidth : public VarReal {
 public:
	VarBandwidth(const char* name, int readonly, double* val)
		: VarReal(name, readonly, val) { }
	virtual void set(const char*);
};

class VarTime : public VarReal {
 public:
	VarTime(const char* name, int readonly, double* val)
		: VarReal(name, readonly, val) { }
	virtual void set(const char*);
};

class VarInt : public Var {
 public:
	VarInt(const char* name, int readonly, int* val);
	virtual void set(const char*);
	virtual const char* get();
 protected:
	int* val_;
};

class VarBool : public VarInt {
 public:
	VarBool(const char* name, int readonly, int* val);
	virtual void set(const char*);
};

/*XXX*/
NsObject::NsObject() : var_(0), trace_(0)
{
}

NsObject::~NsObject()
{
	for (Var* p = var_; p != 0; ) {
		Var* n = p->next_;
		delete p;
		p = n;
	}
}

void NsObject::insert(Var* p)
{
	p->next_ = var_;
	var_ = p;
}

int NsObject::command(int argc, const char*const* argv)
{
	if (argc == 4 && strcmp(argv[1], "set") == 0)
		return (set(argv[2], argv[3]));
	else if (argc == 3) {
		if (strcmp(argv[1], "set") == 0 ||
		     strcmp(argv[1], "get") == 0)
			return (get(argv[2]));
		if (strcmp(argv[1], "trace") == 0) {
			trace_ = (Trace*)TclObject::lookup(argv[2]);
			/*XXX*/
			if (trace_ == 0) {
				Tcl& tcl = Tcl::instance();
				tcl.resultf("ns object trace: no trace %s",
					    argv[2]);
				return (TCL_ERROR);
			}
			return (TCL_OK);
		}
	}
	return (TclObject::command(argc, argv));
}

Var* NsObject::varlookup(const char* var) const
{
	for (Var* p = var_; p != 0; p = p->next_)
		if (strcmp(p->name(), var) == 0)
			return (p);
	Tcl& tcl = Tcl::instance();
	tcl.resultf("ns: no such var: %s", var);
	return (0);
}

int NsObject::get(const char* var)
{
	Var* v = varlookup(var);
	if (v == 0)
		return (TCL_ERROR);
	Tcl::instance().result(v->get());
	return (TCL_OK);
}

int NsObject::set(const char* var, const char* val)
{
	Var* v = varlookup(var);
	if (v == 0)
		return (TCL_ERROR);
	if (v->readonly()) {
		Tcl::instance().resultf("ns: %s is read-only", var);
		return (TCL_ERROR);
	}
	v->set(val);
	return (TCL_OK);
}

void NsObject::link_real(const char* a, const char* name,
		       double* val, int readOnly)
{
	VarReal* v = new VarReal(name, readOnly, val);
	if (a != 0)
		v->set(Tcl::instance().rds(a, name));
	insert(v);
}

void NsObject::link_bw(const char* a, const char* name,
		       double* val, int readOnly)
{
	VarBandwidth* v = new VarBandwidth(name, readOnly, val);
	if (a != 0)
		v->set(Tcl::instance().rds(a, name));
	insert(v);
}

void NsObject::link_time(const char* a, const char* name,
		       double* val, int readOnly)
{
	VarTime* v = new VarTime(name, readOnly, val);
	if (a != 0)
		v->set(Tcl::instance().rds(a, name));
	insert(v);
}

void NsObject::link_int(const char* a, const char* name,
		      int* val, int readOnly)
{
	VarInt* v = new VarInt(name, readOnly, val);
	if (a != 0)
		v->set(Tcl::instance().rds(a, name));
	insert(v);
}

void NsObject::link_bool(const char* a, const char* name,
		      int* val, int readOnly)
{
	VarBool* v = new VarBool(name, readOnly, val);
	if (a != 0)
		v->set(Tcl::instance().rds(a, name));
	insert(v);
}

double
bw_atof(const char* s) 
{
	char wrk[32];
	char* cp = wrk;
	while (isdigit(*s) || *s == 'e' || *s == '+' ||
	       *s == '-' || *s == '.')
		*cp++ = *s++;
	*cp = 0;
	double v = atof(wrk);
	switch (s[0]) {
	case 'k':
	case 'K':
		v *= 1e3;
		break;
	case 'm':
	case 'M':
		v *= 1e6;
		break;
	}
	if (s[1] == 'B')
		v *= 8;

	return (v);
}

double
time_atof(const char* s)
{
	char wrk[32];
	char* cp = wrk;
	while (isdigit(*s) || *s == 'e' || *s == '+' || *s == '-' || *s == '.')
		*cp++ = *s++;
	*cp = 0;
	double v = atof(wrk);
	switch (*s) {
	case 'm':
		v *= 1e-3;
		break;
	case 'u':
		v *= 1e-6;
		break;
	case 'n':
		v *= 1e-9;
		break;
	case 'p':
		v *= 1e-12;
		break;
	}
	return (v);
}

Var::Var(const char* name, int readonly) : readonly_(readonly)
{
	int n = strlen(name) + 1;
	name_ = new char[n];
	strcpy(name_, name);
}

Var::~Var()
{
	delete name_;
}

VarReal::VarReal(const char* name, int readonly, double* val)
	: Var(name, readonly), val_(val)
{
}

const char* VarReal::get()
{
	sprintf(wrk_, "%g", *val_);
	return (wrk_);
}

void VarReal::set(const char* s)
{
	*val_ = atof(s);
}

void VarBandwidth::set(const char* s)
{
	*val_ = bw_atof(s);
}

void VarTime::set(const char* s)
{
	*val_ = time_atof(s);
}

VarInt::VarInt(const char* name, int readonly, int* val)
	: Var(name, readonly), val_(val)
{
}

const char* VarInt::get()
{
	sprintf(wrk_, "%d", *val_);
	return (wrk_);
}

void VarInt::set(const char* s)
{
	*val_ = atoi(s);
}

VarBool::VarBool(const char* name, int readonly, int* val)
	: VarInt(name, readonly, val)
{
}

void VarBool::set(const char* s)
{
	if (isdigit(*s))
		*val_ = atoi(s);
	switch (*s) {
	case 't':
	case 'T':
		*val_ =  1;
		return;
	}
	*val_ = 0;
}
